# -*- coding: utf-8 -*-

"""
Program that copies pdb + mtz files generated with easypipe script,
+ cif and pdb of ligand,
from each dataset in folders with the name of the dataset,
in order to launch Pandda.
The name of the dataset should be 'DIR-PDB TAG', that completed by 'pdb_style' given to PanDDA should reconstitute
the name of the pdb!
"""
#--------------------------------------------------------------
pandda_info = \
"""
For information on PanDDA, see https://pandda.bitbucket.io/

If PanDDA is installed and you have at least 40 datasets, go in 'PANDDA' folder and launch something like: 
pandda.analyse data_dirs='./data/*' pdb_style='*_ok.pdb' cpus=6 checks.all_data_are_valid_values=None 
checks.low_resolution_completeness=None

When done :
% cd pandda
% pandda.inspect
=> properly position the ligand, merge, save
next ...
At the end: show summary, then quit.

Then, change into the directory containing the pandda folder and run the command to export:
% cd ..
% pandda.export pandda_dir='./pandda'
=> copies in a pandda-export folder all built models
"""

#--------------------------Imports-----------------------------
import datetime
from pathlib import Path
import shutil
import os

# My modules
from modules.utils import verbose
import config


#----------------------- Functions-------------
def data_for_pandda(data_dir, res_dirs):
    """
    'pandda' subcommand
    'launch' subcommand should have been launched before
    :param data_dir: data directory
    :param res_dirs: list of results directories names (in 'easYPipe' folder)
    :return: log file
    """

    date = datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")

    """Folder where we work"""
    work_dir = Path.cwd()
    """The existence of following path has been checked in easypipe"""
    phenix_dir = work_dir.joinpath('easYPipe')
    pandda_master_dir = work_dir.joinpath('PANDDA')
    pandda_master_dir.mkdir(exist_ok=True)  # créer un dossier 'PANDDA'

    """Create a log file"""
    nom_log = "pandda-data_easypipe_" + date + ".log"
    pandda_log = verbose.log_start(pandda_master_dir, 'logs', nom_log, verbose.easypipe_logo)

    for res_dir_name in res_dirs:
        res_dir = phenix_dir.joinpath(res_dir_name)
        verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs,
                                                   "Result directory: " + res_dir.name), pandda_log)

        """Create a folder 'PANDDA' and a sub-folder 'data' where data will be copied"""
        pandda_dir = pandda_master_dir.joinpath('PANDDA_' + res_dir_name.split('RESULTS_')[1] )
        pandda_dir.mkdir(exist_ok=True)  # créer un dossier 'PANDDA_' où sera lancé pandda
        pandda_data = pandda_dir.joinpath('data')
        pandda_data.mkdir(exist_ok=True)

        """List of datasets"""
        datasets = [d for d in res_dir.iterdir() if d.is_dir() and d.name != '_mtz_treated']
        verbose.print_and_write("Number of datasets: " + str(len(datasets)), pandda_log)
        verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), pandda_log)

        """Looking for mtz and pdb files"""
        data_OK = 0
        for d in datasets:
            verbose.print_and_write("Dataset: " + d.name, pandda_log)
            mtz_to_copy = [f for f in d.iterdir() if f.is_file() and f.suffix == '.mtz' and f.stem.endswith('_full')]
            if not mtz_to_copy:  # si pas de fichier 'full', recherche 'fast'
                mtz_to_copy = [f for f in d.iterdir() if
                               f.is_file() and f.suffix == '.mtz' and f.stem.endswith('_fast')]
            if mtz_to_copy:
                if len(mtz_to_copy) > 1:
                    # cas où plusieurs mtz traités: garde le plus ancien, i.e. celui traité en premier, i.e. celui avec la meilleure complétude
                    time, mtz = min((mtz.stat().st_mtime, mtz) for mtz in mtz_to_copy)
                else:
                    mtz = mtz_to_copy[0]

                # création du dossier dataset i.e. ce qui est avant .mtz
                dataname = mtz.stem
                verbose.print_and_write("Creation of dataset folder: " + dataname, pandda_log)
                pandda_data.joinpath(dataname).mkdir(exist_ok=True)
                verbose.print_and_write("Copy of data files:", pandda_log)
                verbose.print_and_write(mtz.name, pandda_log)
                mtz_copy = mtz.stem + "_ok.mtz"
                shutil.copyfile(str(mtz), str(pandda_data.joinpath(dataname, mtz_copy)))
                # vérifier que le pdb correspondant existe et le copier
                pdb = d.joinpath(mtz.stem + '.pdb')
                if pdb.exists():
                    verbose.print_and_write(pdb.name, pandda_log)
                    pdb_copy = pdb.stem + "_ok.pdb"
                    shutil.copyfile(str(pdb), str(pandda_data.joinpath(dataname, pdb_copy)))

                # regarder si fichier cif dans dossier ligand
                ligand_to_copy = []
                try:
                    ligand_to_copy = [l for l in d.joinpath('ligand').iterdir() if l.is_file() and l.suffix == '.cif']
                except FileNotFoundError as e:
                    verbose.print_and_write("No ligand file:", pandda_log)
                    verbose.print_and_write(str(e), pandda_log)
                if ligand_to_copy:
                    verbose.print_and_write("Copy of ligand files:", pandda_log)
                    # vérifier que le pdb correspondant au cif existe
                    for lig in ligand_to_copy:
                        if lig.parent.joinpath(lig.stem + '.pdb').exists():
                            verbose.print_and_write(lig.name, pandda_log)
                            shutil.copyfile(str(lig),
                                            str(pandda_data.joinpath(dataname, lig.name)))
                            verbose.print_and_write(lig.stem + '.pdb', pandda_log)
                            shutil.copyfile(str(lig.parent.joinpath(lig.stem + '.pdb')),
                                            str(pandda_data.joinpath(dataname, lig.stem + '.pdb')))
            data_OK += 1
            verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), pandda_log)
        verbose.print_and_write("Number of copied datasets: " + str(data_OK), pandda_log)
        verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs,
                                                               "PanDDA"), pandda_log)
        verbose.print_and_write(pandda_info, pandda_log)
        os.chdir(work_dir)
        return pandda_log












