# -*- coding: utf-8 -*-

"""'reindex' subcommand """

# ----------------------- Imports --------------------------------
import sys
import os
from pathlib import Path
import datetime
import time
import subprocess
import csv
import shutil

# My modules
import config
from modules.utils import verbose
from modules.utils import xtriage
from modules.utils import proteins
from modules.utils import software

#--------------------------Functions------------------------------
# Setup Phenix
if os.name == 'posix':
    env_phenix = software.shell_source(Path(config.PHENIX_PATH).joinpath("phenix_env.sh"))  # sourcer Phenix sous Linux
    shell_choice = False
elif os.name == 'nt':
    env_phenix = software.setup_bat(Path(config.PHENIX_PATH).joinpath("phenix_env.bat"))  # sourcer Phenix sous Windows
    shell_choice = True

def reindex_phenix(mtz, sg_ref, new_mtz_file, logname):
    """
    Try to reindex a mtz file according to a reference space group.
    Write a new mtz file in the same folder as the one of the original mtz file.
    :param mtz: mtz file to reindex
    :param sg_ref: space group of reference
    :param new_mtz_file: mtz file created
    :param logname: log file name with path
    :return: log message
    """
    message = ""
    # TODO: ajouter text=True à la place de encoding='utf-8' quand python >= 3.7
    # des fois bug avec sg (ex : I222) => utiliser le numéro
    args_converter = ['phenix.reflection_file_converter' , '--change_to_space_group='
                      + str(proteins.find_key(proteins.sg_nb_to_symbol, sg_ref)), mtz, '--mtz=' + new_mtz_file]
    reflection_file_converter = subprocess.Popen(args_converter,
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                 shell=shell_choice, env = env_phenix, encoding='utf-8')
    out, err = reflection_file_converter.communicate()
    # Ecrire le fichier log dans le dossier
    with open(logname, 'w') as log:
        log.write(out)
        log.write(err)
    # TESTER si OK, sinon voir si pas problème de labels
    writing_mtz_file = False    # erreur éventuelle à chercher dans le log

    # tester si 'Writing MTZ file:' dans le log
    with open(logname, 'r') as log:
        lignes = log.readlines()
        for li in reversed(lignes):  # parcourir la liste à l'envers (générateur créé)
            if li.find('Writing MTZ file:') >= 0:
                writing_mtz_file = True
                break
    # si pas 'Writing MTZ file:', tester si pbm de labels
    arg_to_add = ""  #  argument to add in order to run again
    if not writing_mtz_file:
        with open(logname, 'r') as log:
            lignes = log.readlines()
            for li in reversed(lignes):  # parcourir la liste à l'envers (générateur créé)
                lignes = log.readlines()
                if li.find('For example:') >= 0:
                    arg_to_add = li.split('For example:')[1].strip()
                    break

    # relancer si problème de label
    if not writing_mtz_file and arg_to_add:
        # la commande --label="F_grenades ..." ne marche pas, mais --label "F_grenades ..." OUI !!
        # et il faut lister les "F_grenades séparément car d'un bloc ne marche pas ??
        option = arg_to_add.split('=')[0]
        option_val = arg_to_add.split('=')[1].split('"')[1].split(",")
        message = "phenix.reflection_file_converter run again with " + "'" + option + "=" + ",".join(option_val) + "'"
        logname_re = str(Path(logname).parent.joinpath(Path(logname).stem + option + '.log'))
        args_converter.extend([option, ','.join(option_val)])
        reflection_file_converter = subprocess.Popen(args_converter,
                                                     stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                                     shell=shell_choice, env=env_phenix, encoding='utf-8')
        out, err = reflection_file_converter.communicate()
        # Ecrire le fichier log dans le dossier
        with open(logname_re, 'w') as log:
            log.write(out)
            log.write(err)
        # tester si 'Writing MTZ file:' dans le log
        with open(logname_re, 'r') as log:
            lignes = log.readlines()
            for li in reversed(lignes):  # parcourir la liste à l'envers (générateur créé)
                if li.find('Writing MTZ file:') >= 0:
                    writing_mtz_file = True
                    break

    # Si 'Writing MTZ file:' dans le log, tester si pas 'error' dans le fichier (il peut y avoir erreur lors de l'écriture)
    reflection_file_converter_err = False
    if writing_mtz_file:
        # chercher dans le log après labels si existe sinon chercher dans le premier log
        if Path(logname_re).is_file():
            logname = logname_re
        with open(logname, 'r') as log:
            lignes = log.readlines()
            for li in reversed(lignes):  # parcourir la liste à l'envers (générateur créé)
                if li.lower().find('error') >= 0:
                    reflection_file_converter_err = True
                    break

    if reflection_file_converter_err:
        message += "\n!!!!!!!!!!!!!! ERROR: phenix.reflection_file_converter failed ..."
        # ATTENTION dans ce cas pas de fichier créé!!
    else:
        message = "phenix.reflection_file_converter completed with success ..."
    return message


def reindex_pointless(mtz, ref_file, sg_ref, new_mtz_file, pointless_logname):
    """
    Try to reindex a mtz file according to a mtz reference file. If it fails, try with a reference space group.
    Write a new mtz file in the same folder as the one of the original mtz file.
    :param mtz: mtz file to reindex
    :param ref_file: reference mtz file
    :param sg_ref: space group of reference
    :param new_mtz_file: mtz file created by Pointless
    :param pointless_logname: pointless log file name with path
    :return: log message
    """
    message = ""
    # pointless with reference file:
    # TODO: ajouter text=True à la place de encoding='utf-8' quand python >= 3.7
    pointless = subprocess.Popen([config.POINTLESS_EXEC, "hklref", ref_file, "hklin", mtz, "hklout", new_mtz_file],
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                 shell=shell_choice, env = env_ccp4, encoding='utf-8')
    out, err = pointless.communicate()
    # Ecrire le fichier log de pointless dans le dossier
    with open(pointless_logname, 'w') as pointless_log:
        pointless_log.write(out)
        pointless_log.write(err)
    # TESTER si "ERROR" dans le log et relancer avec SPACEGROUP
    pointless_err = False
    with open(pointless_logname, 'r') as pointless_log:
        lignes = pointless_log.readlines()
        for l in lignes:
            if "**** ERROR ****" in l:
                pointless_err = True
                break
    if  pointless_err:
        pointless_logname_sg = str(Path(pointless_logname).parent.joinpath(Path(pointless_logname).stem + '_sg.log'))
        # des fois bug avec sg (ex : I222) => utiliser le numéro
        pointless_input = "spacegroup " + str(proteins.find_key(proteins.sg_nb_to_symbol, sg_ref))
        # TODO: ajouter text=True à la place de encoding='utf-8' quand python >= 3.7
        pointless = subprocess.Popen([config.POINTLESS_EXEC, "hklin", mtz, "hklout", new_mtz_file],
                                     stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE,
                                     shell=shell_choice, env = env_ccp4, encoding='utf-8')
        out, err = pointless.communicate(input=pointless_input)
        with open(pointless_logname_sg, 'w') as pointless_log:
            pointless_log.write(out)
            pointless_log.write(err)
        message = "Pointless launched with space group ..."
        with open(pointless_logname_sg, 'r') as pointless_log:
            lignes = pointless_log.readlines()
            for l in lignes:
                if "**** ERROR ****" in l:
                    message += "\n!!!!!!!!!!!!!! ERROR in pointless either with reference file or space group, " \
                               "see pointless log file ..."
    else:
        message = "Pointless completed with reference file ..."
    return message



def reindex_run(data_dir, args, sg_ref):
    """
    'reindex' subcommand,
    'prep' subcommand should have been launched before
    :param data_dir: data directory
    :param args: args from parser
    :param sg_ref: reference space group
    :return: log file
    """

    date = datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")

    """Folder where we work"""
    work_dir = Path.cwd()
    """The existence of following path (created with 'prep' subcommand) has been checked in easypipe"""
    phenix_dir = work_dir.joinpath('easYPipe')
    prep_dir = phenix_dir.joinpath('1a_prep')
    mtz_treat_csv = prep_dir.joinpath('mtz_to_treat_ALL.csv')
    

    """Creation of reindex directory with space group information"""
    reindex_dir = phenix_dir.joinpath('1b_reindex_' + sg_ref)
    reindex_dir.mkdir(exist_ok=True)

    """Create a log file with reindexation information"""
    reindex_dir.joinpath('logs').mkdir(exist_ok=True)
    nom_reindex_log = "reindex_easypipe_" + date + ".log"
    reindex_log = verbose.log_start(reindex_dir, 'logs', nom_reindex_log, verbose.easypipe_logo)
    verbose.print_and_write("Reference space group: " + sg_ref, reindex_log)
    time.sleep(1)
    verbose.print_and_write("Creation of the directory: " + reindex_dir.name, reindex_log)
    verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), reindex_log)

    """Read mtz information in csv file and copy mtz to reindex in reindex folder"""
    mtz_copy_list = []  # liste des mtz copiés pour traitement, sg correspondant
    verbose.print_and_write(
        verbose.cadre_text(config.motif, config.nb_motifs, "Listing of mtz file to reindex"), reindex_log)
    verbose.print_and_write("Search for information in csv file: 'mtz_to_treat_ALL.csv'", reindex_log)
    verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs), reindex_log)
    verbose.print_and_write(
        "Copy of mtz files to reindex (space group different from reference) in '" + reindex_dir.name +
        "' folder:\n", reindex_log)
    with open(mtz_treat_csv, newline='') as csvfile:
        mtz_csv_reader = csv.DictReader(csvfile)
        for row in mtz_csv_reader:
            # vérifier que le sg est différent de celui de ref
            if row['space group'] != sg_ref: #and not row['mtz file'].endswith("_reindexed_" + row['space group'] + ".mtz"):
                # copie le mtz à traiter dans le dossier reindex, en gardant l'arborescence dataset/mtznb_PROC
                mtz_to_copy = phenix_dir.joinpath('0_processed_datasets', row['dataset'], 'data', row['mtz nb'] + '_'
                                                  + row['process name'], row['mtz file'])
                mtz_copy = reindex_dir.joinpath(row['dataset'], row['mtz nb'] + '_' + row['process name'],
                                                row['mtz file'])
                if not mtz_copy.exists():
                    verbose.print_and_write('/'.join(mtz_copy.parts[len(mtz_copy.parts) - 3:]) + " ... copied",
                                                  reindex_log)
                    reindex_dir.joinpath(row['dataset']).mkdir(exist_ok=True)
                    reindex_dir.joinpath(row['dataset'], row['mtz nb'] + '_' + row['process name']).mkdir(exist_ok=True)
                    shutil.copyfile(str(mtz_to_copy), str(mtz_copy))
                else:
                    verbose.print_and_write(
                        '/'.join(mtz_copy.parts[len(mtz_copy.parts) - 3:]) + " ... already copied!", reindex_log)
                mtz_copy_list.append([mtz_copy, row['space group']])

    """Reindexing"""
    mtz_reindex = []  # liste des noms des mtz réindexés, attention si échec le fichier n'existe pas
    verbose.print_and_write(
        verbose.cadre_text(config.motif, config.nb_motifs, "Reindexing mtz files with phenix.reflection_file_converter"),
        reindex_log)
    for mtz, sg in mtz_copy_list:
        #message_pointless = "Pointless not launched yet ..."
        message_reindex = "phenix.reflection_file_converter not launched yet ..."   # Rem : jamais affiché !!
        os.chdir(str(mtz.parent))  # on va dans le dossier contenant le mtz
        new_mtz_file = mtz.parent.joinpath(mtz.stem + "_reindexed_" + sg_ref + ".mtz")  # nom du fichier mtz ré-indexé
        mtz_reindex.append(new_mtz_file)
        if not new_mtz_file.is_file():  # on ne lance pas si existe déjà
            verbose.print_and_write("phenix.reflection_file_converter launched for: " + "/".join(mtz.parts[-3:]),
                                    reindex_log)
            verbose.print_and_write("Original space group: " + sg, reindex_log)
            #pointless_logname = mtz.parent.joinpath(mtz.stem + "_pointless_" + sg_ref + ".log")
            logname = mtz.parent.joinpath(mtz.stem + "_converter_" + sg_ref + ".log")
            #message_pointless = reindex_pointless(str(mtz), ref_mtz, sg_ref, str(new_mtz_file), str(pointless_logname))
            message_reindex = reindex_phenix(str(mtz), sg_ref, str(new_mtz_file), str(logname))
            verbose.print_and_write(message_reindex, reindex_log)
            verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs),
                                    reindex_log)
        else:
            verbose.print_and_write("phenix.reflection_file_converter already launched for: " + "/".join(mtz.parts[-3:]),
                                    reindex_log)
            verbose.print_and_write("Original space group: " + sg, reindex_log)
            verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs),
                                    reindex_log)

    """Launch Xtriage on reindexed mtz file to get information"""
    mtz_reindex_OK = []  # liste que les fichiers qui existent i.e. Pointless OK et avec SG OK après réindexation
    mtz_reindex_notOK = []  # liste les fichiers mtz pour lesquel le sg n'est pas bon après réindex ou réindex a échoué
    verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs,
                                                           "Getting information from reindexed mtz files with Xtriage"),
                                  reindex_log)
    for mtz in mtz_reindex:
        if mtz.is_file():
            space_group = ""
            nom_xtriage = mtz.parent.joinpath("xtriage_" + mtz.stem + ".log")
            xtriage_verbose = mtz.parent.joinpath("xtriage-verbose.log")
            if not nom_xtriage.is_file():   # si log xtriage existe, ne le relance pas
                verbose.print_and_write("Xtriage launched for: " + "/".join(mtz.parts[-3:]), reindex_log)
                xtriage.xtriage_mtz(mtz, xtriage_verbose, nom_xtriage)
            # si xtriage échoue, pas de fichier log (Ex : Sorry: Data are unusually incomplete (9.9%);
            # this usually indicates a corrupted input file or data processing errors.)
            if nom_xtriage.is_file():
                space_group = xtriage.sg_in_xtriage_log(nom_xtriage)
            if space_group == sg_ref:
                mtz_reindex_OK.append(mtz)
            else:
                mtz_reindex_notOK.append(mtz)
        else:
            mtz_reindex_notOK.append(mtz)

    """Modification of mtz_to_treat_ALL.csv file and copy of reindexed mtz in processed_datasets/reindex"""
    # ne le modifie que si nécessaire (pointless Ok et new SG OK)
    if mtz_reindex_OK:
        verbose.print_and_write(
            verbose.cadre_text(config.motif, config.nb_motifs, "Modification of mtz_to_treat_ALL.csv file "
                                    "to mtz_to_treat_ALL_reindexed.csv file and copy of reindexed mtz "), reindex_log)
        # Renommage du csv d'origine pour sauvegarde
        #mtz_treat_csv_ori = reindex_dir.joinpath(mtz_treat_csv.name + ".old")
        #verbose.print_and_write("Changing name of csv file before reindexing: " + mtz_treat_csv_ori.name,
        #                              reindex_log)
        # verbose.print_and_write("Moving csv file before reindexing to '" + reindex_dir.name + "' folder", reindex_log)
        #os.rename(str(mtz_treat_csv), str(mtz_treat_csv_ori))
        #verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs),
        #                        reindex_log)


        # Création d'un nouveau csv modifié (nom du mtz et complétude)
        mtz_treat_csv_reindex = reindex_dir.joinpath(mtz_treat_csv.stem + "_reindexed.csv")
        verbose.print_and_write("Creating new " + mtz_treat_csv_reindex.name + " file ... will be stored in "
                                + reindex_dir.name + " folder", reindex_log)
        verbose.print_and_write("Updating mtz file name and completeness when necessary:", reindex_log)
        with open(mtz_treat_csv, newline='') as csvfile_ori:
            mtz_csv_reader = csv.DictReader(csvfile_ori)
            header = mtz_csv_reader.fieldnames  # récupère le header
            with open(mtz_treat_csv_reindex, 'w', newline='') as csvfile:
                csv_writer = csv.writer(csvfile, quoting=csv.QUOTE_MINIMAL)
                csv_writer.writerow(header)  # header
                for row in mtz_csv_reader:
                    mtz_diff = False  # est-ce que le mtz a été réindexé i.e est différent de l'origine
                    for mtz in mtz_reindex_OK:
                        # chercher le mtz à remplacer en comparant le nom du dossier parent
                        if mtz.parts[-3] == row['dataset'] and mtz.parts[-2] == row['mtz nb'] + "_" + row[
                            'process name']:
                            for fi in mtz.parent.iterdir():
                                # Récupération des infos dans le log Xtriage
                                if fi.is_file and fi.name.startswith("xtriage_"):
                                    completeness = xtriage.completeness_in_xtriage_log(fi)
                                    space_group = xtriage.sg_in_xtriage_log(fi)
                                    # calcul de la résolution et unit cell
                                    # (cas où échec xtriage à l'étape prep => pas de résolution à copier dans mtz_treat_csv)
                                    resolution = xtriage.reso_in_xtriage_log(fi)
                                    unitcell = xtriage.unitcell_in_xtriage_log(fi)
                                    # copie du log xtriage du mtz réindexé dans '0_processed_datasets'
                                    xtriage_new = phenix_dir.joinpath('0_processed_datasets', row['dataset'], 'data',
                                                          row['mtz nb'] + "_" + row['process name'], fi.name)
                                    shutil.copyfile(fi, xtriage_new)

                            verbose.print_and_write("\nDataset: " + row['dataset'], reindex_log)
                            verbose.print_and_write("New mtz for " + row['mtz nb'] + ": " + mtz.name, reindex_log)
                            verbose.print_and_write("New completeness (%): " + str(completeness), reindex_log)
                            verbose.print_and_write("Original space group: " + row['space group'], reindex_log)
                            verbose.print_and_write("New space group: " + str(space_group), reindex_log)
                            csv_writer.writerow(list(row.values())[0:3] + [mtz.name] + [resolution]
                                                + [completeness] + [space_group] + [unitcell] + list(row.values())[8:])
                            # copie du mtz réindexé dans '0_processed_datasets'
                            verbose.print_and_write("Copy of new mtz in '0_processed_datasets'", reindex_log)
                            mtz_new = phenix_dir.joinpath('0_processed_datasets', row['dataset'], 'data',
                                                          row['mtz nb'] + "_" + row['process name'], mtz.name)
                            shutil.copyfile(mtz, mtz_new)
                            verbose.print_and_write("xtriage logfile of new mtz als copied in '0_processed_datasets'",
                                                    reindex_log)

                            mtz_diff = True
                            break
                    if not mtz_diff:
                        csv_writer.writerow(list(row.values()))
    else:
        verbose.print_and_write(verbose.chaine_xfois(config.sub_motif, config.nb_motifs),
                                reindex_log)
        verbose.print_and_write("No mtz reindexed successfully ...", reindex_log)

    """Summary"""
    verbose.print_and_write(verbose.chaine_xfois(config.motif, config.nb_motifs), reindex_log)
    verbose.print_and_write(verbose.easypipe_logo, reindex_log)
    verbose.print_and_write(verbose.cadre_text(config.motif, config.nb_motifs, "SUMMARY "),
                            reindex_log)
    verbose.print_and_write("mtz candidates for reindexing (space group different from reference file): "
                                  + str(len(mtz_copy_list)) + "\n", reindex_log)
    for mtz, sg in mtz_copy_list:
        verbose.print_and_write("/".join(mtz.parts[-3:]) + " / " + sg, reindex_log)
    verbose.print_and_write("\nNumber of successfully reindexed mtz: " + str(len(mtz_reindex_OK)), reindex_log)
    if mtz_reindex_OK:  # cas où il y a eu des réindexation qui ont marché
        for mtz in mtz_reindex_OK:
            verbose.print_and_write("/".join(mtz.parts[-3:]), reindex_log)
        verbose.print_and_write("\nYou can visualize new " + mtz_treat_csv_reindex.name + " file after reindexing in "
                                + reindex_dir.name + " folder.", reindex_log)
    if mtz_reindex_notOK:
        verbose.print_and_write("\nNumber of not successfully reindexed mtz: " + str(len(mtz_reindex_notOK)), reindex_log)
        for mtz in mtz_reindex_notOK:
            for mtz_cop , sg in mtz_copy_list:
                if "/".join(mtz.parts[-3:-1]) == "/".join(mtz_cop.parts[-3:-1]):
                    verbose.print_and_write("/".join(mtz.parts[-3:-1]) + " / " + sg, reindex_log)
                    break
    if args.subcommands == 'auto':
        verbose.print_and_write(
            verbose.cadre_text(config.motif, config.nb_motifs, "Next step in automatic mode: subcommand 'launch' ..."),
            reindex_log)
    os.chdir(work_dir)
    return reindex_log

